//## begin module%1.4%.codegen_version preserve=yes
//   Read the documentation to learn more about C++ code generator
//   versioning.
//## end module%1.4%.codegen_version

//## begin module%3FBB8933012C.cm preserve=no
//	  %X% %Q% %Z% %W%
//## end module%3FBB8933012C.cm

//## begin module%3FBB8933012C.cp preserve=no
//## end module%3FBB8933012C.cp

//## Module: WmapRegion%3FBB8933012C; Package body
//## Subsystem: XSModel::Model::MixFunction%3F74624902C0
//## Source file: /data/gordon/headas/Xspec/src/XSModel/Model/MixFunction/WmapRegion.cxx

//## begin module%3FBB8933012C.additionalIncludes preserve=no
//## end module%3FBB8933012C.additionalIncludes

//## begin module%3FBB8933012C.includes preserve=yes
#include <CCfits/CCfits>
#include <CCfits/FITSBase.h>
#include <sstream>
//## end module%3FBB8933012C.includes

// WmapRegion
#include "WmapRegion.h"
//## begin module%3FBB8933012C.declarations preserve=no
//## end module%3FBB8933012C.declarations

//## begin module%3FBB8933012C.additionalDeclarations preserve=yes
//## end module%3FBB8933012C.additionalDeclarations


// Class WmapRegion::WmapRegionError 

WmapRegion::WmapRegionError::WmapRegionError (const string& msg)
  //## begin WmapRegion::WmapRegionError::WmapRegionError%3FBB972A00C8.hasinit preserve=yes
  : YellowAlert("Wmap region error: ")
  //## end WmapRegion::WmapRegionError::WmapRegionError%3FBB972A00C8.hasinit
  //## begin WmapRegion::WmapRegionError::WmapRegionError%3FBB972A00C8.initialization preserve=yes
  //## end WmapRegion::WmapRegionError::WmapRegionError%3FBB972A00C8.initialization
{
  //## begin WmapRegion::WmapRegionError::WmapRegionError%3FBB972A00C8.body preserve=yes
  std::cerr << msg << std::endl;
  //## end WmapRegion::WmapRegionError::WmapRegionError%3FBB972A00C8.body
}


// Class WmapRegion 

WmapRegion::WmapRegion()
  //## begin WmapRegion::WmapRegion%3FBB89E40133_const.hasinit preserve=yes
  : m_lowerLeft(std::pair<Real,Real>(.0,.0)),
    m_upperRight(std::pair<Real,Real>(.0,.0)),
    m_nx(0),
    m_ny(0),
    m_wmap(),
    m_wmrbn(1),
    m_file(0),
    m_refBinKeys(),
    m_wmrbnKey(""),
    m_version(.0),
    m_wmapPos(),
    m_isPixOn()
  //## end WmapRegion::WmapRegion%3FBB89E40133_const.hasinit
  //## begin WmapRegion::WmapRegion%3FBB89E40133_const.initialization preserve=yes
  //## end WmapRegion::WmapRegion%3FBB89E40133_const.initialization
{
  //## begin WmapRegion::WmapRegion%3FBB89E40133_const.body preserve=yes
  //## end WmapRegion::WmapRegion%3FBB89E40133_const.body
}



//## Other Operations (implementation)
void WmapRegion::readCommonKeys ()
{
  //## begin WmapRegion::readCommonKeys%3FBB924702CC.body preserve=yes
  CCfits::PHDU& hdu = m_file->pHDU();
  const CCfits::PHDU& chdu = const_cast<const CCfits::PHDU&>(hdu);
  const string fileName = chdu.parent()->name();
  if (hdu.axes() == 2)
  {
     m_nx = static_cast<int>(hdu.axis(0));
     m_ny = static_cast<int>(hdu.axis(1));
  }
  else
  {
     string msg("Primary HDU does not have 2 axes in file: ");
     msg += fileName;
     throw WmapRegionError(msg);        
  }
  hdu.readAllKeys();
  
  string errorKey("HDUVERS");
  try
  {
     m_version = getVersion();
     if (m_version < 0)
     {
        string msg("Error reading ");
        msg += errorKey + string(" keyword in file: ");
        msg += fileName;
        throw WmapRegionError(msg);
     }
     else if (m_version < 2)
     {
        m_wmrbnKey = "WMREBIN";
        m_refBinKeys.first = "CRVAL1";
        m_refBinKeys.second = "CRVAL2";
     }
     else
     {
        m_wmrbnKey = "CDELT1P";
        m_refBinKeys.first = "CRVAL1P";
        m_refBinKeys.second = "CRVAL2P";
     }
     errorKey = m_wmrbnKey;
     getWmrbn();
     errorKey = m_refBinKeys.first + ", " + m_refBinKeys.second;
     getCorners();
  }
  catch (CCfits::FitsException&)
  {
     string msg("Error reading ");
     msg += errorKey + string(" keyword(s) in file: ");
     msg += fileName;
     throw WmapRegionError(msg);
  }
  //## end WmapRegion::readCommonKeys%3FBB924702CC.body
}

void WmapRegion::readWmap (int* nullVal)
{
  //## begin WmapRegion::readWmap%3FBB946A0366.body preserve=yes
  CCfits::PHDU& primary = m_file->pHDU();
  try
  {
     int nPix = m_nx * m_ny;
     std::valarray<int> tmpArray;
     m_file->pHDU().read(tmpArray, 1, nPix, nullVal);
     
     const int EMPTY = -1;
     m_wmapPos.clear();
     m_isPixOn.resize(nPix, false);
     for (int i=0; i<nPix; ++i)
     {
        if (tmpArray[i] != EMPTY)
        {
           m_isPixOn[i] = true;
           m_wmapPos.push_back(i);
        }
     }
     int sz = m_wmapPos.size();
     m_wmap.resize(sz);
     // SunCC v6 has linktime trouble when m_isPixOn is used
     // as a mask_array.  So, do this the more cumbersome way...
     for (int i=0; i<sz; ++i)
        m_wmap[i] = tmpArray[m_wmapPos[i]];
  }
  catch (CCfits::FitsException&)
  {
     string msg("Error attempting to read image in primary HDU of ");
     const CCfits::PHDU& chdu = const_cast<const CCfits::PHDU&>(primary);
     msg += chdu.parent()->name();
     throw WmapRegionError(msg);     
  }
  //## end WmapRegion::readWmap%3FBB946A0366.body
}

Real WmapRegion::getVersion ()
{
  //## begin WmapRegion::getVersion%3FBBB45E034D.body preserve=yes
  string keyword("HDUVERS");
  string keyval("");
  Real vers = -1.0;
  CCfits::PHDU& primary = m_file->pHDU();
  try
  {
     primary.keyWord(keyword).value(keyval);
  }
  catch (CCfits::FitsException&)
  {
     // Failed once, try slightly different keyword.
     keyword += "1";
     primary.keyWord(keyword).value(keyval);
  }
  std::istringstream ssVal(keyval);
  // If this fails (keyval does not start with a Real), vers
  // will remain negative and calling program will deal with it.
  ssVal >> vers;
  
  return vers;
  //## end WmapRegion::getVersion%3FBBB45E034D.body
}

void WmapRegion::getWmrbn ()
{
  //## begin WmapRegion::getWmrbn%3FBBB47E01E4.body preserve=yes
  float tmpFloat = .0;
  m_file->pHDU().keyWord(m_wmrbnKey).value(tmpFloat);
  m_wmrbn = int(tmpFloat);
  //## end WmapRegion::getWmrbn%3FBBB47E01E4.body
}

void WmapRegion::getCorners ()
{
  //## begin WmapRegion::getCorners%3FBBB49A007D.body preserve=yes
  float tmpFloat = .0;
  m_file->pHDU().keyWord(m_refBinKeys.first).value(tmpFloat);
  m_lowerLeft.first = static_cast<Real>(tmpFloat);
  m_file->pHDU().keyWord(m_refBinKeys.second).value(tmpFloat);
  m_lowerLeft.second = static_cast<Real>(tmpFloat);
  m_upperRight.first = m_lowerLeft.first + static_cast<Real>(m_nx*m_wmrbn);
  m_upperRight.second = m_lowerLeft.second + static_cast<Real>(m_ny*m_wmrbn);
  //## end WmapRegion::getCorners%3FBBB49A007D.body
}

void WmapRegion::read (const string& fileName, int* nullVal)
{
  //## begin WmapRegion::read%3FBCCAF70004.body preserve=yes
  // Re-open the spectrum file as given by fileName, which should
  // be the full path name.
  try
  {
     std::auto_ptr<CCfits::FITS> pf(new CCfits::FITS(fileName));
     m_file = pf.get();
     // Any CCfits exceptions thrown inside the following
     // read functions will be caught there and rethrown
     // as Xspec exceptions.
     readCommonKeys();
     readSpecificKeys();
     readWmap(nullVal);
     // m_file pointer will be invalid once pf goes out of scope.
     m_file = 0;
  }
  catch (CCfits::FITS::CantOpen)
  {
     string msg = "Cannot open file: ";
     msg += fileName;
     msg += "\nfor retrieving data for mixing model.";
     throw WmapRegionError(msg);
  }
  catch (CCfits::FitsException&)
  {
     throw WmapRegionError("");
  }
  //## end WmapRegion::read%3FBCCAF70004.body
}

// Additional Declarations
  //## begin WmapRegion%3FBB89E40133.declarations preserve=yes
  WmapRegion::~WmapRegion() {}
  //## end WmapRegion%3FBB89E40133.declarations
//## begin module%3FBB8933012C.epilog preserve=yes
//## end module%3FBB8933012C.epilog
